"""Use Monte Carlo to propagate uncertainties"""

import numpy as np
from multiprocessing import Pool
import warnings

'''___Authorship___'''
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

class MCPropagation:
    def __init__(self,steps,parallel_cores=0):
        """
        Initialise MC Propagator

        :param steps: number of MC iterations
        :type steps: int
        """

        self.MCsteps = steps
        self.parallel_cores = parallel_cores
        if parallel_cores>1:
            self.pool=Pool(parallel_cores)

    def propagate_random(self,func,x,u_x,cov_x=None,param_fixed=None,corr_between=None,return_corr=False,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1):
        """
        Propagate random uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Random uncertainties arise when there is no correlation between repeated measurements.
        It is possible (though rare) that there is a correlation in one of the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of random uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param cov_x: list of covariance matrices (n,n) along non-repeating axis, defaults to None
        :type cov_x: list[array], optional
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        yshape,u_x,repeat_axis,repeat_dims,corr_axis = self.perform_checks(func,x,u_x,
                                                                           repeat_dims,
                                                                           corr_axis,
                                                                           output_vars)
        if repeat_axis >= 0:
            n_repeats=yshape[repeat_axis]
            outs = np.empty(n_repeats,dtype=object)
            for i in range(n_repeats):
                xb, u_xb = self.select_repeated_x(x,u_x,param_fixed,i,repeat_axis,n_repeats)
                outs[i] = self.propagate_random(func,xb,u_xb,cov_x,param_fixed,corr_between,return_corr,
                                                return_samples,repeat_dims,corr_axis=corr_axis,output_vars=output_vars)
            return self.combine_repeated_outs(outs,yshape,n_repeats,len(x),repeat_axis,return_corr,return_samples,output_vars)

        else:
            MC_data = np.empty(len(x),dtype=np.ndarray)
            for i in range(len(x)):
                if cov_x is None or cov_x[i]=="rand":
                    MC_data[i] = self.generate_samples_random(x[i],u_x[i])
                elif cov_x[i]=="syst":
                    MC_data[i] = self.generate_samples_systematic(x[i],u_x[i])
                else:
                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),cov_x[i]).reshape(x[i].shape+(self.MCsteps,))

                #print(MC_data[i].nbytes)
            if corr_between is not None:
                MC_data = self.correlate_samples_corr(MC_data,corr_between)
            return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars)

    def propagate_systematic(self,func,x,u_x,cov_x=None,param_fixed=None,corr_between=None,return_corr=False,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1):
        """
        Propagate systematic uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Systematic uncertainties arise when there is full correlation between repeated measurements.
        There is a often also a correlation between measurements along the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param cov_x: list of covariance matrices (n,n) along non-repeating axis, defaults to None
        :type cov_x: list[array], optional
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        yshape,u_x,repeat_axis,repeat_dims,corr_axis = self.perform_checks(func,x,u_x,
                                                                       repeat_dims,
                                                                       corr_axis,
                                                                       output_vars)

        if repeat_axis >= 0:
            n_repeats=yshape[repeat_axis]
            outs = np.empty(n_repeats,dtype=object)
            for i in range(n_repeats):
                xb,u_xb = self.select_repeated_x(x,u_x,param_fixed,i,repeat_axis,n_repeats)
                outs[i] = self.propagate_systematic(func,xb,u_xb,cov_x,param_fixed,corr_between,
                                                return_corr,return_samples,repeat_dims,
                                                corr_axis=corr_axis,
                                                output_vars=output_vars)
            return self.combine_repeated_outs(outs,yshape,n_repeats,len(x),repeat_axis,
                                              return_corr,return_samples,output_vars)

        else:
            MC_data = np.empty(len(x),dtype=np.ndarray)
            for i in range(len(x)):
                if u_x[i] is None:
                    u_x[i] = np.zeros_like(x[i])

                if cov_x is None or cov_x[i] == "syst":
                    MC_data[i] = self.generate_samples_systematic(x[i],u_x[i])
                elif cov_x[i] == "rand":
                    MC_data[i] = self.generate_samples_random(x[i],u_x[i])
                else:
                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),cov_x[i]).reshape(x[i].shape+(self.MCsteps,))

            if corr_between is not None:
                MC_data = self.correlate_samples_corr(MC_data,corr_between)

            return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars)

    def propagate_cov(self,func,x,cov_x,param_fixed=None,corr_between=None,return_corr=True,return_samples=False,repeat_dims=-99,corr_axis=-99,output_vars=1):
        """
        Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        The covariance matrix can represent the full covariance matrix between all measurements in all dimensions.
        Alternatively if there are repeated measurements specified in repeat_dims, the covariance matrix is given
        for the covariance along the dimension that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param cov_x: list of covariance matrices on input quantities (usually numpy arrays). In case the input quantity is an array of shape (m,o), the covariance matrix needs to be given as an array of shape (m*o,m*o).
        :type cov_x: list[array]
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        yshape,u_x,repeat_axis,repeat_dims,corr_axis = self.perform_checks(func,x,cov_x,
                                                                           repeat_dims,
                                                                           corr_axis,
                                                                           output_vars)

        if repeat_axis >= 0:
            n_repeats=yshape[repeat_axis]
            outs = np.empty(n_repeats,dtype=object)
            for i in range(n_repeats):
                xb,_ = self.select_repeated_x(x,x,param_fixed,i,repeat_axis,n_repeats)
                outs[i] = self.propagate_cov(func,xb,cov_x,param_fixed,corr_between,
                                                return_corr,return_samples,repeat_dims,
                                                corr_axis=corr_axis,
                                                output_vars=output_vars)
            return self.combine_repeated_outs(outs,yshape,n_repeats,len(x),repeat_axis,
                                              return_corr,return_samples,output_vars)
        else:
            MC_data = np.empty(len(x),dtype=np.ndarray)
            for i in range(len(x)):
                if not hasattr(x[i],"__len__"):
                    MC_data[i] = self.generate_samples_systematic(x[i],cov_x[i])
                elif (all((cov_x[i]==0).flatten())): #This is the case if one of the variables has no uncertainty
                    MC_data[i] = np.tile(x[i].flatten(),(self.MCsteps,1)).T
                else:
                    MC_data[i] = self.generate_samples_cov(x[i].flatten(),cov_x[i]).reshape(x[i].shape+(self.MCsteps,))
            if corr_between is not None:
                MC_data = self.correlate_samples_corr(MC_data,corr_between)

        return self.process_samples(func,MC_data,return_corr,return_samples,corr_axis,output_vars)

    def perform_checks(self,func,x,u_x,repeat_dims,corr_axis,output_vars):
        if output_vars == 1:
            yshape = np.array(func(*x)).shape
        else:
            yshape = np.array(func(*x)[0]).shape

        if x[0].shape != yshape and self.parallel_cores == 0:
            print(
                "It looks like one of your input quantities is not the same shape as the measurand."
                "This is not a problem, but means you cannot use array operations in your measurement function."
                "You will need to set parallel_cores to 1 or higher when creating your MCPropagation object.")
            exit()

        for i in range(len(x)):
            if u_x[i] is None:
                u_x[i] = np.zeros(x[i].shape)

        if isinstance(repeat_dims,int):
            repeat_axis = repeat_dims
            repeat_dims = -99
        else:
            repeat_axis = repeat_dims[0]
            repeat_dims = repeat_dims[1]
            if repeat_axis<repeat_dims:
                repeat_dims -= 1

        if repeat_axis >= 0:
            if corr_axis > repeat_axis:
                corr_axis -= 1
            elif corr_axis == repeat_axis:
                print("corr_axis and repeat_axis keywords should not be the same.")
                exit()

        return yshape,u_x,repeat_axis,repeat_dims,corr_axis

    def select_repeated_x(self,x,u_x,param_fixed,i,repeat_axis,n_repeats):
        xb=np.zeros(len(x),dtype=object)
        u_xb=np.zeros(len(u_x),dtype=object)
        for j in range(len(x)):
            if len(x[j].shape) > repeat_axis:
                if (x[j].shape[repeat_axis]!=n_repeats):
                    xb[j] = x[j]
                    u_xb[j] = u_x[j]
                elif param_fixed is not None:
                    if param_fixed[j]==True:
                        xb[j] = x[j]
                        u_xb[j] = u_x[j]
                elif repeat_axis == 0 :
                    xb[j]=x[j][i]
                    u_xb[j] = u_x[j][i]
                elif repeat_axis == 1:
                    xb[j] = x[j][:,i]
                    u_xb[j] = u_x[j][:,i]
                elif repeat_axis == 2:
                    xb[j] = x[j][:,:,i]
                    u_xb[j] = u_x[j][:,:,i]
                else:
                    warnings.warn("The repeat axis is too large to be dealt with by the"
                                  "current version of punpy.")
            else:
                if (len(x[j])==n_repeats):
                    xb[j] = x[j][i]
                    u_xb[j] = u_x[j][i]
                    if param_fixed is not None:
                        if param_fixed[j]==True:
                            xb[j] = x[j]
                            u_xb[j] = u_x[j]
                else:
                    xb[j] = x[j]
                    u_xb[j] = u_x[j]
        return xb, u_xb

    def combine_repeated_outs(self,outs,yshape,n_repeats,lenx,repeat_axis,return_corr,return_samples,output_vars):

        if not return_corr and not return_samples:
            if output_vars == 1:
                u_func = np.zeros(yshape)
                if repeat_axis == 0:
                    for i in range(len(outs)):
                        u_func[i] = outs[i]
                elif repeat_axis == 1:
                    for i in range(len(outs)):
                        u_func[:,i] = outs[i]
                elif repeat_axis == 2:
                    for i in range(len(outs)):
                        u_func[:,:,i] = outs[i]
            else:
                u_func = np.zeros((output_vars,)+yshape)
                if repeat_axis == 0:
                    for i in range(len(outs)):
                        u_func[:,i] = outs[i]
                elif repeat_axis == 1:
                    for i in range(len(outs)):
                        u_func[:,:,i] = np.array(outs[i])
                elif repeat_axis == 2:
                    for i in range(len(outs)):
                        u_func[:,:,:,i] = outs[i]
            return u_func

        else:
            if output_vars == 1:
                u_func = np.zeros(yshape)
                if repeat_axis == 0:
                    for i in range(len(outs)):
                        u_func[i] = outs[i][0]
                elif repeat_axis == 1:
                    for i in range(len(outs)):
                        u_func[:,i] = outs[i][0]
                elif repeat_axis == 2:
                    for i in range(len(outs)):
                        u_func[:,:,i] = outs[i][0]
            else:
                u_func = np.zeros((output_vars,)+yshape)
                if repeat_axis == 0:
                    for i in range(len(outs)):
                        u_func[:,i] = outs[i][0]
                elif repeat_axis == 1:
                    for i in range(len(outs)):
                        u_func[:,:,i] = outs[i][0]
                elif repeat_axis == 2:
                    for i in range(len(outs)):
                        u_func[:,:,:,i] = outs[i][0]

            returns = np.empty(len(outs[0]),dtype=object)
            returns[0]=u_func
            extra_index=0
            if return_corr:
                returns[1]=np.mean([outs[i][1] for i in range(n_repeats)],axis=0)
                extra_index+=1

            if output_vars>1:
                returns[1+extra_index]=np.mean([outs[i][1+extra_index] for i in range(n_repeats)],axis=0)
                extra_index+=1

            if return_samples:
                returns[1+extra_index] = [np.vstack([outs[i][1+extra_index][k] for i in range(n_repeats)]) for k in range(lenx)]
                returns[2+extra_index] = [np.vstack([outs[i][2+extra_index][k] for i in range(n_repeats)]) for k in range(lenx)]

            return returns

    def process_samples(self,func,data,return_corr,return_samples,corr_axis=-99,output_vars=1):
        """
        Run the MC-generated samples of input quantities through the measurement function and calculate
        correlation matrix if required.

        :param func: measurement function
        :type func: function
        :param data: MC-generated samples of input quantities
        :type data: array[array]
        :param return_corr: set to True to return correlation matrix of measurand
        :type return_corr: bool
        :param return_samples: set to True to return generated samples
        :type return_samples: bool
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        if self.parallel_cores==0:
            MC_y = np.array(func(*data))
        elif self.parallel_cores==1:
            # In order to Process the MC iterations separately, the array with the input quantities has to be reordered
            # so that it has the same length (i.e. the first dimension) as the number of MCsteps.
            # First we move the axis with the same length as self.MCsteps from the last dimension to the fist dimension
            data2 = [np.moveaxis(dat,-1,0) for dat in data]
            # The function can then be applied to each of these MCsteps
            MC_y2 = list(map(func,*data2))
            # We then reorder to bring it back to the original shape
            MC_y = np.moveaxis(MC_y2,0,-1)

        else:
            # We again need to reorder the input quantities samples in order to be able to pass them to p.starmap
            # We here use lists to iterate over and order them slightly different as the case above.
            data2=[[data[j][...,i] for j in range(len(data))] for i in range(self.MCsteps)]
            MC_y2=np.array(self.pool.starmap(func,data2))
            MC_y = np.moveaxis(MC_y2,0,-1)

        u_func = np.std(MC_y,axis=-1)
        if not return_corr:
            if return_samples:
                return u_func,MC_y,data
            else:
                return u_func
        else:
            if output_vars==1:
                corr_y = self.calculate_corr(MC_y,corr_axis)
                if return_samples:
                    return u_func,corr_y,MC_y,data
                else:
                    return u_func,corr_y

            else:
                #create an empty arrays and then populate it with the correlation matrix for each output parameter individually
                corr_ys=np.empty(output_vars,dtype=object)
                for i in range(output_vars):
                    corr_ys[i] = self.calculate_corr(MC_y[i],corr_axis)

                #calculate correlation matrix between the different outputs produced by the measurement function.
                corr_out=np.corrcoef(MC_y.reshape((output_vars,-1)))

                if return_samples:
                    return u_func,corr_ys,corr_out,MC_y,data
                else:
                    return u_func,corr_ys,corr_out

    def calculate_corr(self,MC_y,corr_axis=-99):
        """
        Calculate the correlation matrix between the MC-generated samples of output quantities.
        If corr_axis is specified, this axis will be the one used to calculate the correlation matrix (e.g. if corr_axis=0 and x.shape[0]=n, the correlation matrix will have shape (n,n)).
        This will be done for each combination of parameters in the other dimensions and the resulting correlation matrices are averaged.

        :param MC_y: MC-generated samples of the output quantities (measurands)
        :type MC_y: array
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :return: correlation matrix
        :rtype: array
        """
        #print("the shape is:",MC_y.shape)

        if len(MC_y.shape) <3:
                corr_y = np.corrcoef(MC_y)

        elif len(MC_y.shape) == 3:
            if corr_axis==0:
                corr_ys = np.empty(len(MC_y[0]),dtype=object)
                for i in range(len(MC_y[0])):
                    corr_ys[i] = np.corrcoef(MC_y[:,i])
                corr_y = np.mean(corr_ys,axis=0)

            elif corr_axis==1:
                corr_ys = np.empty(len(MC_y),dtype=object)
                for i in range(len(MC_y)):
                    corr_ys[i] = np.corrcoef(MC_y[i])
                corr_y = np.mean(corr_ys,axis=0)

            else:
                MC_y = MC_y.reshape((MC_y.shape[0]*MC_y.shape[1],self.MCsteps))
                corr_y = np.corrcoef(MC_y)

        elif len(MC_y.shape) == 4:
            if corr_axis == 0:
                corr_ys = np.empty(len(MC_y[0])*len(MC_y[0,0]),dtype=object)
                for i in range(len(MC_y[0])):
                    for j in range(len(MC_y[0,0])):
                        corr_ys[i+j*len(MC_y[0])] = np.corrcoef(MC_y[:,i,j])
                corr_y = np.mean(corr_ys,axis=0)

            elif corr_axis == 1:
                corr_ys = np.empty(len(MC_y)*len(MC_y[0,0]),dtype=object)
                for i in range(len(MC_y)):
                    for j in range(len(MC_y[0,0])):
                        corr_ys[i+j*len(MC_y)] = np.corrcoef(MC_y[i,:,j])
                corr_y = np.mean(corr_ys,axis=0)

            elif corr_axis == 2:
                corr_ys = np.empty(len(MC_y)*len(MC_y[0]),dtype=object)
                for i in range(len(MC_y)):
                    for j in range(len(MC_y[0])):
                        corr_ys[i+j*len(MC_y)] = np.corrcoef(MC_y[i,j])
                corr_y = np.mean(corr_ys,axis=0)
            else:
                MC_y = MC_y.reshape((MC_y.shape[0]*MC_y.shape[1]*MC_y.shape[2],self.MCsteps))
                corr_y = np.corrcoef(MC_y)
        else:
            print("MC_y has too high dimensions. Reduce the dimensionality of the input data")
            exit()

        return corr_y

    def generate_samples_random(self,param,u_param):
        """
        Generate MC samples of input quantity with random (Gaussian) uncertainties.

        :param param: values of input quantity (mean of distribution)
        :type param: float or array
        :param u_param: uncertainties on input quantity (std of distribution)
        :type u_param: float or array
        :return: generated samples
        :rtype: array
        """
        if not hasattr(param,"__len__"):
            return np.random.normal(size=self.MCsteps)*u_param+param
        elif len(param.shape) == 1:
            return np.random.normal(size=(len(param),self.MCsteps))*u_param[:,None]+param[:,None]
        elif len(param.shape) == 2:
            return np.random.normal(size=param.shape+(self.MCsteps,))*u_param[:,:,None]+param[:,:,None]
        elif len(param.shape) == 3:
            return np.random.normal(size=param.shape+(self.MCsteps,))*u_param[:,:,:,None]+param[:,:,:,None]
        else:
            print("parameter shape not supported")
            exit()


    def generate_samples_systematic(self,param,u_param):
        """
        Generate correlated MC samples of input quantity with systematic (Gaussian) uncertainties.

        :param param: values of input quantity (mean of distribution)
        :type param: float or array
        :param u_param: uncertainties on input quantity (std of distribution)
        :type u_param: float or array
        :return: generated samples
        :rtype: array
        """
        if not hasattr(param,"__len__"):
            return np.random.normal(size=self.MCsteps)*u_param+param
        elif len(param.shape) == 1:
            return np.dot(u_param[:,None],np.random.normal(size=self.MCsteps)[None,:])+param[:,None]
        elif len(param.shape) == 2:
            return np.dot(u_param[:,:,None],np.random.normal(size=self.MCsteps)[:,None,None])[:,:,:,0]+param[:,:,None]
        elif len(param.shape) == 3:
            return np.dot(u_param[:,:,:,None],np.random.normal(size=self.MCsteps)[:,None,None,None])[:,:,:,:,0,0]+param[:,:,:,None]
        else:
            print("parameter shape not supported")
            exit()

    def generate_samples_cov(self,param,cov_param):
        """
        Generate correlated MC samples of input quantity with a given covariance matrix.
        Samples are generated independent and then correlated using Cholesky decomposition.

        :param param: values of input quantity (mean of distribution)
        :type param: array
        :param cov_param: covariance matrix for input quantity
        :type cov_param: array
        :return: generated samples
        :rtype: array
        """
        try:
            L = np.linalg.cholesky(cov_param)
        except:
            L = self.nearestPD_cholesky(cov_param)

        return np.dot(L,np.random.normal(size=(len(param),self.MCsteps)))+param[:,None]

    def correlate_samples_corr(self,samples,corr):
        """
        Method to correlate independent samples of input quantities using correlation matrix and Cholesky decomposition.

        :param samples: independent samples of input quantities
        :type samples: array[array]
        :param corr: correlation matrix between input quantities
        :type corr: array
        :return: correlated samples of input quantities
        :rtype: array[array]
        """
        if np.max(corr) > 1 or len(corr) != len(samples):
            raise ValueError("The correlation matrix between variables is not the right shape or has elements >1.")
        else:
            try:
                L = np.array(np.linalg.cholesky(corr))
            except:
                L = self.nearestPD_cholesky(corr)

            #Cholesky needs to be applied to Gaussian distributions with mean=0 and std=1,
            #We first calculate the mean and std for each input quantity
            means = np.array([np.mean(samples[i]) for i in range(len(samples))])
            stds = np.array([np.std(samples[i]) for i in range(len(samples))])

            #We normalise the samples with the mean and std, then apply Cholesky, and finally reapply the mean and std.
            if all(stds!=0):
                return np.dot(L,(samples-means)/stds)*stds+means

            #If any of the variables has no uncertainty, the normalisation will fail. Instead we leave the parameters without uncertainty unchanged.
            else:
                samples_out=samples[:]
                id_nonzero=np.where(stds!=0)
                samples_out[id_nonzero]=np.dot(L[id_nonzero][:,id_nonzero],(samples[id_nonzero]-means[id_nonzero])/stds[id_nonzero])[:,0]*stds[id_nonzero]+means[id_nonzero]
                return samples_out

    @staticmethod
    def nearestPD_cholesky(A):
        """
        Find the nearest positive-definite matrix

        :param A: correlation matrix or covariance matrix
        :type A: array
        :return: nearest positive-definite matrix
        :rtype: array

        Copied and adapted from [1] under BSD license.
        A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [2], which
        credits [3].
        [1] https://gist.github.com/fasiha/fdb5cec2054e6f1c6ae35476045a0bbd
        [2] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd
        [3] N.J. Higham, "Computing a nearest symmetric positive semidefinite
        matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
        """

        B = (A+A.T)/2
        _,s,V = np.linalg.svd(B)

        H = np.dot(V.T,np.dot(np.diag(s),V))

        A2 = (B+H)/2

        A3 = (A2+A2.T)/2

        try:
            return np.linalg.cholesky(A3)
        except:

            spacing = np.spacing(np.linalg.norm(A))

            I = np.eye(A.shape[0])
            k = 1
            while not MCPropagation.isPD(A3):
                mineig = np.min(np.real(np.linalg.eigvals(A3)))
                A3 += I*(-mineig*k**2+spacing)
                k += 1

            if np.any(abs(A-A3)/(A+0.0001) > 0.0001):
                raise ValueError(
                    "One of the provided covariance matrix is not postive definite. Covariance matrices need to be at least positive semi-definite. Please check your covariance matrix.")
            else:
                warnings.warn(
                    "One of the provided covariance matrix is not positive definite. It has been slightly changed (less than 0.01% in any element) to accomodate our method.")
                return np.linalg.cholesky(A3)

    @staticmethod
    def isPD(B):
        """
        Returns true when input is positive-definite, via Cholesky

        :param B: matrix
        :type B: array
        :return: true when input is positive-definite
        :rtype: bool
        """
        try:
            _ = np.linalg.cholesky(B)
            return True
        except np.linalg.LinAlgError:
            return False



from punpy.mc.mc_propagation import MCPropagation
from punpy.jacobian.jacobian_propagation import JacobianPropagation
from punpy.mc.MCMC_retrieval import MCMCRetrieval
from punpy.jacobian.jacobian_retrieval import JacobianRetrieval

from ._version import get_versions

__author__ = "Sam Hunt <sam.hunt@npl.co.uk>"
__all__ = []

__version__ = get_versions()["version"]
del get_versions



"""Run MCMC for retrieval"""

'''___Built-In Modules___'''

'''___Third-Party Modules___'''
import numpy as np
import emcee
from multiprocessing import Pool
import time
import os

os.environ["OMP_NUM_THREADS"] = "1"

'''___NPL Modules___'''

'''___Authorship___'''
__author__ = "Pieter De Vis"
__created__ = "01/03/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

inds_cache = {}


# lock = threading.Lock()
class MCMCRetrieval:
    def __init__(self,measurement_function,observed,syst_uncertainty=None,rand_uncertainty=None,cov=None,parallel_cores=4,uplims=+np.inf,downlims=-np.inf):
        self.measurement_function=measurement_function
        self.observed = observed
        self.rand_uncertainty = np.array([rand_uncertainty])
        self.syst_uncertainty = np.array([syst_uncertainty])
        if cov is None:
            self.invcov = cov
        else:
            self.invcov = np.linalg.inv(np.ascontiguousarray(cov))
        self.uplims = np.array(uplims)
        self.downlims = np.array(downlims)
        self.parallel_cores=parallel_cores

    def run_retrieval(self,theta_0,nwalkers,steps,burn_in,return_samples=True,return_corr=False):
        # if self.syst_uncertainty is not None:
        #     theta_0=np.append([0.],theta_0)
        ndimw = len(theta_0)
        pos = [theta_0*np.random.normal(1.0,0.1,theta_0.shape)+np.random.normal(0.0,0.001,theta_0.shape) for i in
               range(nwalkers)]
        #print(self.measurement_function(theta_0))
        p = Pool(self.parallel_cores)
        sampler = emcee.EnsembleSampler(nwalkers,ndimw,self.lnprob,pool=p)
        sampler.run_mcmc(pos,steps,progress=False)

        samples = sampler.chain[:,:,:].reshape((-1,ndimw))[burn_in::]
        medians = np.median(samples,axis=0)
        unc_up = (np.percentile(samples,84,axis=0)-medians)
        unc_down = -(np.percentile(samples,16,axis=0)-medians)
        unc_avg= (unc_up+unc_down)/2.
        corr=(np.corrcoef(samples.T))


        if return_samples:
            if return_corr:
                return medians, unc_avg, corr, samples
            else:
                return medians,unc_avg,samples

        else:
            if return_corr:
                return medians, unc_avg, corr
            else:
                return medians,unc_avg

    def find_chisum(self,theta):
        model=self.measurement_function(theta)
        diff = model-self.observed-theta[0]
        if np.isfinite(np.sum(diff)):
            if self.invcov is None:
                return np.sum((diff)**2/self.rand_uncertainty**2)
            else:
                #print(diff,np.linalg.inv(self.cov),np.dot(np.dot(diff.T,self.invcov),diff))
                return np.dot(np.dot(diff.T,self.invcov),diff)
        else:
            return np.inf

    # def upper_triangular_to_symmetric(self,ut):
    #     n = ut.shape[0]
    #     try:
    #         inds = inds_cache[n]
    #     except KeyError:
    #         inds = np.tri(n, k=-1, dtype=np.bool)
    #         inds_cache[n] = inds
    #     ut[inds] = ut.T[inds]

    # def fast_positive_definite_inverse(self,m):
    #     cholesky, info = lapack.dpotrf(m)
    #     if info != 0:
    #         raise ValueError('dpotrf failed on input {}'.format(m))
    #     inv, info = lapack.dpotri(cholesky)
    #     if info != 0:
    #         raise ValueError('dpotri failed on input {}'.format(cholesky))
    #     self.upper_triangular_to_symmetric(inv)
    #     return inv

    def lnlike(self,theta):
        #print(theta,self.find_chisum(theta))
        return -0.5*(self.find_chisum(theta))

    def lnprior(self,theta):
        if all(self.downlims < theta[1::]) and all(self.uplims > theta[1::]):
            #if self.syst_uncertainty[0] is None:
                return 0
            # else:
            #     return -0.5*(theta[0]**2/self.syst_uncertainty**2)
        else:
            return -np.inf

    def lnprob(self,theta):
        lp_prior = self.lnprior(theta)
        if not np.isfinite(lp_prior):
            return -np.inf
        lp = self.lnlike(theta)
        # if lp < -99999999999:
        #     return -np.inf
        return lp_prior+lp



"""Use Monte Carlo to propagate uncertainties"""

import numpy as np
import numdifftools as nd
import warnings

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def calculate_Jacobian(fun,x):
    """
    Calculate the local Jacobian of function y=f(x) for a given value of x

    :param fun: measurement function
    :type fun: function
    :param x: local values of input quantities
    :type x: array
    :return: Jacobian
    :rtype: array
    """
    Jfun = nd.Jacobian(fun)
    Jx = Jfun(x.flatten())
    if len(Jx)!=len(fun(x).flatten()):
        warnings.warn("Dimensions of the Jacobian were flipped because its shape didn't match "
              "the shape of the output of the function. (probably because there was "
              "only 1 input qty)")
        Jx=Jx.T

    #print("Jacobian calculation took:",time.time()-timeb,timeb-timea)
    return Jx

def convert_corr_to_cov(corr, u):
    """
    Convert correlation matrix to covariance matrix

    :param corr: correlation matrix
    :type corr: array
    :param u: uncertainties
    :type u: array
    :return: covariance matrix
    :rtype: array
    """
    return u.reshape((-1, 1)) * corr * (u.reshape((1, -1)))

def convert_cov_to_corr(cov, u):
    """
    Convert covariance matrix to correlation matrix

    :param corr: covariance matrix
    :type corr: array
    :param u: uncertainties
    :type u: array
    :return: correlation matrix
    :rtype: array
    """
    return 1 / u.reshape((-1, 1)) * cov / (u.reshape((1, -1)))



"""
Tests for mc propagation class
"""

import unittest
import numpy as np
import numpy.testing as npt
import punpy.utilities.utilities as util
from punpy.jacobian.jacobian_propagation import JacobianPropagation

'''___Authorship___'''
__author__ = "Pieter De Vis"
__created__ = "14/4/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

def function(x1,x2):
    return x1**2 - 10*x2

x1=np.ones(200)*10
x2=np.ones(200)*30
x1err=np.ones(200)
x2err=2*np.ones(200)

xs=np.array([x1,x2])
xerrs=np.array([x1err,x2err])

#below, the higher order Taylor expansion terms have been taken into account, and amount to 2.
yerr_uncorr=802**0.5*np.ones(200)
yerr_corr=2**0.5*np.ones(200)

def functionb(x1,x2):
    return 2* x1 - x2

x1b=np.ones((20,3))*50
x2b=np.ones((20,3))*30
x1errb=np.ones((20,3))
x2errb=2*np.ones((20,3))

xsb=np.array([x1b,x2b])
xerrsb=np.array([x1errb,x2errb])

yerr_uncorrb=8**0.5*np.ones((20,3))
yerr_corrb=np.zeros((20,3))

def functionc(x1,x2,x3):
    return x1  +4*x2 -2*x3

x1c=np.ones(200)*10
x2c=np.ones(200)*10
x3c=np.ones(200)*10

x1errc=12*np.ones(200)
x2errc=5*np.ones(200)
x3errc=np.zeros(200)

xsc=np.array([x1c,x2c,x3c])
xerrsc=np.array([x1errc,x2errc,x3errc])
corr_c=np.array([[1,0.9999999,0],[0.99999999,1.,0],[0.,0.,1.]])
yerr_uncorrc=544**0.5*np.ones(200)
yerr_corrc=1024**0.5*np.ones(200)

def functiond(x1,x2):
    return 2* x1 - x2, 2*x1+x2

x1d=np.ones((20,3,4))*50
x2d=np.ones((20,3,4))*30
x1errd=np.ones((20,3,4))
x2errd=2*np.ones((20,3,4))

xsd=np.array([x1d,x2d])
xerrsd=np.array([x1errd,x2errd])
corr_d=np.ones((2,2))#np.array([[1,0.9999999,0.9999999],[0.99999999,1.,0.99999999],[0.9999999,0.9999999,1.]])

yerr_uncorrd=[8**0.5*np.ones((20,3,4)),8**0.5*np.ones((20,3,4))]
yerr_corrd=[np.zeros((20,3,4)),16**0.5*np.ones((20,3,4))]


class TestMCPropagation(unittest.TestCase):
    """
    Class for unit tests
    """
    def test_propagate_random(self):
        prop = JacobianPropagation()

        uf,ucorr = prop.propagate_random(function,xs,xerrs,return_corr=True)
        npt.assert_allclose(ucorr,np.eye(len(ucorr)),atol=0.06)
        npt.assert_allclose(uf,yerr_uncorr,rtol=0.06)
        ucov= util.convert_corr_to_cov(ucorr,uf)
        ucorr2= util.convert_cov_to_corr(ucov,uf)
        npt.assert_allclose(ucorr,ucorr2,atol=0.01)

        uf = prop.propagate_random(function,xs,xerrs,corr_between=np.ones((2,2)))
        npt.assert_allclose(uf,np.zeros_like(uf),atol=0.01)

        #b
        ufb,ucorrb = prop.propagate_random(functionb,xsb,xerrsb,return_corr=True)
        npt.assert_allclose(ucorrb,np.eye(len(ucorrb)),atol=0.06)
        npt.assert_allclose(ufb,yerr_uncorrb,rtol=0.06)

        ufb = prop.propagate_random(functionb,xsb,xerrsb,corr_between=np.ones((2,2)))
        npt.assert_allclose(ufb,yerr_corrb,atol=0.03)

        #c
        ufc,ucorrc = prop.propagate_random(functionc,xsc,xerrsc,return_corr=True)
        npt.assert_allclose(ucorrc,np.eye(len(ucorrc)),atol=0.06)
        npt.assert_allclose(ufc,yerr_uncorrc,rtol=0.06)

        ufc = prop.propagate_random(functionc,xsc,xerrsc,corr_between=corr_c)
        npt.assert_allclose(ufc,yerr_corrc,rtol=0.06)

        #d
        ufd,ucorrd = prop.propagate_random(functiond,xsd,xerrsd,return_corr=True,output_vars=2)
        npt.assert_allclose(ucorrd[0],np.eye(len(ucorrd[0])),atol=0.06)
        npt.assert_allclose(ufd,yerr_uncorrd,rtol=0.06)

        ufd = prop.propagate_random(functiond,xsd,xerrsd,corr_between=corr_d,output_vars=2)
        npt.assert_allclose(ufd[0],yerr_corrd[0],atol=0.06)
        npt.assert_allclose(ufd[1],yerr_corrd[1],rtol=0.06)


    # def test_propagate_systematic(self):
    #     prop = MCPropagation(30000,parallel_cores=3)
    #
    #     uf,ucorr = prop.propagate_systematic(function,xs,xerrs,return_corr=True)
    #     npt.assert_allclose(ucorr,np.ones_like(ucorr),atol=0.06)
    #     npt.assert_allclose(uf,yerr_uncorr,rtol=0.06)
    #
    #     uf = prop.propagate_systematic(function,xs,xerrs,
    #                                corr_between=np.ones((2,2)))
    #     npt.assert_allclose(uf,yerr_corr,rtol=0.06)
    #
    #     #b
    #     ufb,ucorrb = prop.propagate_systematic(functionb,xsb,xerrsb,return_corr=True)
    #     npt.assert_allclose(ucorrb,np.ones_like(ucorrb),atol=0.06)
    #     npt.assert_allclose(ufb,yerr_uncorrb,rtol=0.06)
    #
    #     ufb,ucorrb = prop.propagate_systematic(functionb,xsb,xerrsb,return_corr=True,corr_axis=0)
    #     npt.assert_allclose(ucorrb,np.ones_like(ucorrb),atol=0.06)
    #
    #     ufb,ucorrb = prop.propagate_systematic(functionb,xsb,xerrsb,return_corr=True,corr_axis=1)
    #     npt.assert_allclose(ucorrb,np.ones_like(ucorrb),atol=0.06)
    #
    #     ufb = prop.propagate_systematic(functionb,xsb,xerrsb,corr_between=np.ones((2,2)))
    #     npt.assert_allclose(ufb,yerr_corrb,atol=0.03)
    #
    #     #c
    #     ufc,ucorrc = prop.propagate_systematic(functionc,xsc,xerrsc,return_corr=True)
    #     npt.assert_allclose(ucorrc,np.ones_like(ucorrc),atol=0.06)
    #     npt.assert_allclose(ufc,yerr_uncorrc,rtol=0.06)
    #
    #     ufc = prop.propagate_systematic(functionc,xsc,xerrsc,corr_between=corr_c)
    #     npt.assert_allclose(ufc,yerr_corrc,rtol=0.06)
    #
    #     ufd,ucorrd,corr_out = prop.propagate_systematic(functiond,xsd,xerrsd,return_corr=True,corr_axis=0,
    #                                                     output_vars=2)
    #     npt.assert_allclose(ucorrd[0],np.ones_like(ucorrd[0]),atol=0.06)
    #
    #     ufd,ucorrd,corr_out = prop.propagate_systematic(functiond,xsd,xerrsd,return_corr=True,corr_axis=1,
    #                                                     output_vars=2)
    #     npt.assert_allclose(ucorrd[1],np.ones_like(ucorrd[1]),atol=0.06)
    #
    #     ufd,ucorrd,corr_out = prop.propagate_systematic(functiond,xsd,xerrsd,return_corr=True,corr_axis=2,
    #                                                     output_vars=2)
    #     npt.assert_allclose(ucorrd[0],np.ones_like(ucorrd[0]),atol=0.06)
    #
    # def test_propagate_cov(self):
    #     prop = MCPropagation(20000)
    #
    #     cov = [util.convert_corr_to_cov(np.eye(len(xerr.flatten())),xerr) for xerr in xerrs]
    #     uf,ucorr = prop.propagate_cov(function,xs,cov,return_corr=True)
    #     npt.assert_allclose(ucorr,np.eye(len(ucorr)),atol=0.06)
    #     npt.assert_allclose(uf,yerr_uncorr,rtol=0.06)
    #
    #     cov = [util.convert_corr_to_cov(np.ones((len(xerr.flatten()),len(xerr.flatten())))+np.eye(len(xerr)),xerr) for xerr in xerrs]
    #     uf,ucorr = prop.propagate_cov(function,xs,cov,return_corr=True)
    #     npt.assert_allclose(uf,yerr_uncorr*2**0.5,rtol=0.06)
    #
    #     cov = [util.convert_corr_to_cov(np.eye(len(xerr.flatten())),xerr) for xerr in xerrs]
    #     uf,ucorr = prop.propagate_cov(function,xs,cov,return_corr=True,corr_between=np.ones((2,2)))
    #     npt.assert_allclose(ucorr,np.eye(len(ucorr)),atol=0.06)
    #     npt.assert_allclose(uf,yerr_corr,rtol=0.06)
    #
    #     #b
    #     covb = [util.convert_corr_to_cov(np.eye(len(xerrb.flatten())),xerrb) for xerrb in xerrsb]
    #     ufb,ucorrb = prop.propagate_cov(functionb,xsb,covb,return_corr=True)
    #     npt.assert_allclose(ucorrb,np.eye(len(ucorrb)),atol=0.06)
    #     npt.assert_allclose(ufb,yerr_uncorrb,rtol=0.06)
    #
    #     covb = [util.convert_corr_to_cov(np.ones((len(xerrb.flatten()),len(xerrb.flatten())))+np.eye(len(xerrb.flatten())),xerrb) for xerrb in
    #            xerrsb]
    #     ufb,ucorrb = prop.propagate_cov(functionb,xsb,covb,return_corr=True)
    #     npt.assert_allclose(ufb,yerr_uncorrb*2**0.5,rtol=0.06)
    #
    #     covb = [util.convert_corr_to_cov(np.eye(len(xerrb.flatten())),xerrb) for xerrb in xerrsb]
    #     ufb,ucorrb = prop.propagate_cov(functionb,xsb,covb,return_corr=True,corr_between=np.ones((2,2)))
    #     npt.assert_allclose(ucorrb,np.eye(len(ucorrb)),atol=0.06)
    #     npt.assert_allclose(ufb,yerr_corrb,atol=0.03)
    #
    #     #c
    #     covc = [util.convert_corr_to_cov(np.eye(len(xerrc.flatten())),xerrc) for xerrc in xerrsc]
    #     ufc,ucorrc = prop.propagate_cov(functionc,xsc,covc,return_corr=True)
    #     npt.assert_allclose(ucorrc,np.eye(len(ucorrc)),atol=0.06)
    #     npt.assert_allclose(ufc,yerr_uncorrc,rtol=0.06)
    #
    #     covc = [util.convert_corr_to_cov(np.ones((len(xerrc.flatten()),len(xerrc.flatten())))+np.eye(len(xerrc)),xerrc) for xerrc in
    #             xerrsc]
    #     ufc,ucorrc = prop.propagate_cov(functionc,xsc,covc,return_corr=True)
    #     npt.assert_allclose(ufc,yerr_uncorrc*2**0.5,rtol=0.06)
    #
    #     covc = [util.convert_corr_to_cov(np.eye(len(xerrc.flatten())),xerrc) for xerrc in xerrsc]
    #     ufc,ucorrc = prop.propagate_cov(functionc,xsc,covc,return_corr=True,corr_between=corr_c)
    #     npt.assert_allclose(ucorrc,np.eye(len(ucorrc)),atol=0.06)
    #     npt.assert_allclose(ufc,yerr_corrc,rtol=0.06)

if __name__ == '__main__':
    unittest.main()



"""Use Monte Carlo to propagate uncertainties"""

import numpy as np
from multiprocessing import Pool
import numdifftools as nd
import punpy.utilities.utilities as util

'''___Authorship___'''
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

class JacobianPropagation:
    def __init__(self,parallel_cores=0):
        """
        Initialise MC Propagator

        :param steps: number of MC iterations
        :type steps: int
        """

        self.parallel_cores = parallel_cores

    def propagate_random(self,func,x,u_x,corr_between=None,return_corr=False,corr_axis=-99,output_vars=1):
        """
        Propagate random uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors or images.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of random uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        yshape=np.shape(func(*x))
        if output_vars==1:
            fun = lambda c: func(*c.reshape(len(x),-1))
        else:
            fun = lambda c: np.concatenate(func(*c.reshape(len(x),-1)))
        Jx=util.calculate_Jacobian(fun,x)
        if corr_between is None:
            corr_x = np.eye(len(x.flatten()))
        else:
            corrs=[np.eye(len(xi.flatten())) for xi in x]
            corr_x=self.calculate_flattened_corr(corrs,corr_between)
        cov_x=util.convert_corr_to_cov(corr_x,u_x)
        return self.process_jacobian(Jx,cov_x,yshape,return_corr,corr_axis,output_vars)

    def calculate_flattened_corr(self,corrs,corr_between):
        totcorrlen=0
        for i in range(len(corrs)):
            totcorrlen += len(corrs[i])
        totcorr=np.eye(totcorrlen)
        for i in range(len(corrs)):
            for j in range(len(corrs)):
                totcorr[i*len(corrs[i]):(i+1)*len(corrs[i]),j*len(corrs[j]):(j+1)*len(corrs[j])]=corr_between[i,j]*corrs[i]**0.5*corrs[j]**0.5
        return totcorr

    def process_jacobian(self,J,covx,shape_y,return_corr,corr_axis=-99,output_vars=1):
        covy=np.dot(np.dot(J,covx),J.T)
        u_func=np.diag(covy)**0.5
        corr_y=util.convert_cov_to_corr(covy,u_func)
        #print("test",J,covx,covy)
        if not return_corr:
            return u_func.reshape(shape_y)
        else:
            if output_vars==1:
                return u_func.reshape(shape_y),corr_y
            else:
                #create an empty arrays and then populate it with the correlation matrix for each output parameter individually
                corr_ys=np.empty(output_vars,dtype=object)
                for i in range(output_vars):
                    corr_ys[i] = corr_y[int(i*len(corr_y)/output_vars):
                                        int((i+1)*len(corr_y)/output_vars),
                                        int(i*len(corr_y)/output_vars):
                                        int((i+1)*len(corr_y)/output_vars)]

                # #calculate correlation matrix between the different outputs produced by the measurement function.
                # corr_out=np.corrcoef(MC_y.reshape((output_vars,-1)))

                return u_func.reshape(shape_y),corr_ys#,corr_out

    def propagate_systematic(self,func,x,u_x,corr_between=None,return_corr=False,corr_axis=-99,output_vars=1):
        """
        Propagate systematic uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors or images.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        yshape = np.shape(func(*x))
        if output_vars == 1:
            fun = lambda c:func(*c.reshape(len(x),-1))
        else:
            fun = lambda c:np.concatenate(func(*c.reshape(len(x),-1)))
        Jx = util.calculate_Jacobian(fun,x)
        if corr_between is None:
            corr_x = np.ones((len(x.flatten()),len(x.flatten())))
        else:
            corrs = [np.ones((len(xi.flatten()),len(xi.flatten()))) for xi in x]
            corr_x = self.calculate_flattened_corr(corrs,corr_between)
        cov_x = util.convert_corr_to_cov(corr_x,u_x)
        return self.process_jacobian(Jx,cov_x,yshape,return_corr,corr_axis,output_vars)


    def propagate_both(self,func,x,u_x_rand,u_x_syst,corr_between=None,return_corr=True,corr_axis=-99,output_vars=1):
        """
        Propagate random and systematic uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors or images.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x_rand: list of random uncertainties on input quantities (usually numpy arrays)
        :type u_x_rand: list[array]
        :param u_x_syst: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x_syst: list[array]
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        yshape = np.shape(func(*x))
        if output_vars == 1:
            fun = lambda c:func(*c.reshape(len(x),-1))
        else:
            fun = lambda c:np.concatenate(func(*c.reshape(len(x),-1)))
        Jx = util.calculate_Jacobian(fun,x)
        if corr_between is None:
            corr_x = np.eye(len(x.flatten()))
        else:
            corrs = [np.eye(len(xi.flatten())) for xi in x]
            corr_x = self.calculate_flattened_corr(corrs,corr_between)
        cov_x = util.convert_corr_to_cov(corr_x,u_x)
        return self.process_jacobian(Jx,cov_x,yshape,return_corr,corr_axis,output_vars)

    def propagate_cov(self,func,x,cov_x,corr_between=None,return_corr=True,corr_axis=-99,output_vars=1):
        """
        Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
        Input quantities can be floats, vectors or images.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param cov_x: list of covariance matrices on input quantities (usually numpy arrays). In case the input quantity is an array of shape (m,o), the covariance matrix needs to be given as an array of shape (m*o,m*o).
        :type cov_x: list[array]
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        yshape = np.shape(func(*x))
        if output_vars == 1:
            fun = lambda c:func(*c.reshape(len(x),-1))
        else:
            fun = lambda c:np.concatenate(func(*c.reshape(len(x),-1)))
        Jx = util.calculate_Jacobian(fun,x)
        if corr_between is None:
            corr_between = np.eye(len(x))
        corr_x  = [util.convert_cov_to_corr(corr_x[i],u_x[i]) for i in range(len(x))]
        corr_x_full = self.calculate_flattened_corr(corr_x,corr_between)

        cov_x_full = util.convert_corr_to_cov(corr_x_full,u_x)
        return self.process_jacobian(Jx,cov_x_full,yshape,return_corr,corr_axis,output_vars)



"""Run MCMC for retrieval"""

'''___Built-In Modules___'''

'''___Third-Party Modules___'''
import numpy as np
import emcee
from multiprocessing import Pool
import time
import punpy.utilities.utilities as util
import os
from scipy.optimize import minimize

os.environ["OMP_NUM_THREADS"] = "1"

'''___NPL Modules___'''

'''___Authorship___'''
__author__ = "Pieter De Vis"
__created__ = "01/03/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

inds_cache = {}


# lock = threading.Lock()
class JacobianRetrieval:
    def __init__(self,measurement_function,observed,syst_uncertainty=None,rand_uncertainty=None,cov=None,uplims=+np.inf,downlims=-np.inf):
        self.measurement_function=measurement_function
        self.observed = observed
        self.rand_uncertainty = np.array([rand_uncertainty])
        self.syst_uncertainty = np.array([syst_uncertainty])
        if cov is None:
            self.invcov = cov
        else:
            self.invcov = np.linalg.inv(np.ascontiguousarray(cov))
            #print(observed,cov,self.invcov)

        self.uplims = np.array(uplims)
        self.downlims = np.array(downlims)

    def run_retrieval(self,theta_0,return_corr=True):
        res = minimize(self.find_chisum,theta_0)
        Jx=util.calculate_Jacobian(self.measurement_function,res.x)

        return res.x, self.process_inverse_jacobian(Jx,return_corr)

    def process_inverse_jacobian(self,J,return_corr=True):
        covx = np.linalg.inv(np.dot(np.dot(J.T,self.invcov),J))
        u_func = np.diag(covx)**0.5
        corr_x = util.convert_cov_to_corr(covx,u_func)
        if return_corr:
            return u_func,corr_x
        else:
            return u_func
        # if not return_corr:
        #     return u_func.reshape(shape_y)
        # else:
        #     if output_vars == 1:
        #         return u_func.reshape(shape_y),corr_y
        #     else:
        #         #create an empty arrays and then populate it with the correlation matrix for each output parameter individually
        #         corr_ys = np.empty(output_vars,dtype=object)
        #         for i in range(output_vars):
        #             corr_ys[i] = corr_y[int(i*len(corr_y)/output_vars):
        #                                 int((i+1)*len(corr_y)/output_vars),
        #                          int(i*len(corr_y)/output_vars):
        #                          int((i+1)*len(corr_y)/output_vars)]
        #
        #         # #calculate correlation matrix between the different outputs produced by the measurement function.
        #         # corr_out=np.corrcoef(MC_y.reshape((output_vars,-1)))
        #
        #         return u_func.reshape(shape_y),corr_ys  #,corr_out


    def find_chisum(self,theta):
        model = self.measurement_function(theta)
        diff = model-self.observed-theta[0]
        if np.isfinite(np.sum(diff)):
            if self.invcov is None:
                return np.sum((diff)**2/self.rand_uncertainty**2)
            else:
                #print(diff,np.linalg.inv(self.cov),np.dot(np.dot(diff.T,self.invcov),diff))
                return np.dot(np.dot(diff.T,self.invcov),diff)
        else:
            return np.inf